// ==================================================================
// file: vector_sum.cu
// date: october 2012
// author: Jean-Michel Richer
// email: jean-michel.richer@univ-angers.fr
// ==================================================================

/**
 * link with 
 *   -lrt -lcudart
 * use with (for example) :
 *   --vector-size=10 --max-repeat=10 --method=1
 *
 */
#include <iostream>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <time.h>
#include "vector_sum.h"

using namespace std;



// ==================================================================
// CHRONO
// ==================================================================

struct timespec chrono_start_event;
struct timespec chrono_stop_event;

void start_chrono() {
	clock_gettime( CLOCK_MONOTONIC, &chrono_start_event);
}

void stop_chrono() {
	clock_gettime( CLOCK_MONOTONIC, &chrono_stop_event);
}

double elapsed_time() {
	return double (chrono_stop_event.tv_sec - chrono_start_event.tv_sec) +
		(chrono_stop_event.tv_nsec - chrono_start_event.tv_nsec)/1e9;
}

// ==================================================================

/**
 * report CUDA error if cudaGetLastError() is different from
 * cudaSuccess. An error message will be displayed
 * @param s error message
 */
void cudaReportError(char *s) {
	cudaError_t n = cudaGetLastError();
	if (n != cudaSuccess) {
		cerr << "==========================" << endl;
		cerr << "CUDA ERROR: " << cudaGetErrorString(n) << endl;
		cerr << s << endl;
		cerr << "==========================" << endl;
		exit(EXIT_FAILURE);
	}
}
      
static struct option long_options[] = {
	{"vector-size",  	required_argument, 0, 's'},
	{"max-repeat", 		required_argument, 0, 'r' },
	{"method", 			required_argument, 0, 'm' }, 
	{0, 0, 0, 0}
};

/**
 * main parameters of computation
 *
 */
 // size of vector (please usem ultiple of 16)
int vector_size = 131072;
// number of times the computation is performed
int max_repeat = 100;
// method used
int method_id = 1;

/**
 * main data
 */
// Vector on host
float *vector_h[3];
// Vector on device
float *vector_d[3];

float a_h = 0.5;

// ==================================================================
// METHODS
// ==================================================================

/* see file kernel_cpu.cpp */

void kernel_empty_caller();
void kernel_cpu_caller();
void kernel_cpu_sse_caller();


Method methods[] = {
	{ "no_method", kernel_empty_caller },
	{ "cpu", kernel_cpu_caller },
	{ "cpu_sse", kernel_cpu_sse_caller },
	// mark end of array
	{ "", NULL }
};


int count_methods() {
	int i=1;
	while (methods[i].caller != NULL) {
		++i;
	}
	return i-1;
}

void kernel_empty_caller() { 
	cerr << "Error: define method between 1 and " << count_methods() << endl;
	exit(EXIT_FAILURE);
}

// ==================================================================
// RESOURCES ALLOCATION
// ==================================================================

// in cpu_resources.cpp
void cpu_allocate_resources();
void cpu_free_resources();

/**
 * allocate data on GPU 
 */
void gpu_allocate_resources() {
}

/**
 * free data on GPU
 */
void gpu_free_resources() {
}


/**
 * main program
 */
int main( int argc, char *argv[]) {
	int option_index;
	
	while (1) {
		option_index = 0;
		int c = getopt_long (argc, argv, "s:r:m:", long_options, &option_index);
		if (c == -1) break;
		
        switch(c) {
		case 's':
			vector_size = atoi(optarg);
			break;
		case 'r':
			max_repeat = atoi(optarg);
			break;
		case 'm':
			method_id = atoi(optarg);
			break;
		}
	}
	
	cout << "parameters" << endl;
	cout << "====================" << endl;
	cout << "vector_size=" << vector_size << endl;
	cout << "max_repeat=" << max_repeat << endl;
	cout << "method_id=" << method_id << endl;
	
	cpu_allocate_resources();
	gpu_allocate_resources();
		
	start_chrono();
	
	for (int i=0; i<max_repeat; ++i) {
		methods[method_id].caller();
	}
	
	stop_chrono();
	cout << "elapsed=" << elapsed_time() << endl;
	
	cpu_free_resources();
	gpu_free_resources();
	
	return EXIT_SUCCESS;
}

